from pathlib import Path

# from edkrule.edk_rule import EdkRule
from edkrule.utils.output.printer import Printer


class RuleShow:
    # def __init__(self, rule_string: str):
    #     self._expression = EdkRule().expression(rule_string)

    def output(self, printer: Printer, output_path, data):
        printer.output(output_path, data, Printer.default_jinja_template())



class Character:
    @staticmethod
    def iscolon(char: str):
        return char == ":"

    @staticmethod
    def isalpha(char: str):
        return char.isalpha()

    @staticmethod
    def isdigit(char: str):
        return char.isdigit()

    @staticmethod
    def isplus(char: str): return char == "+"

    @staticmethod
    def isminus(char: str): return char == "-"

    @staticmethod
    def isdivide(char: str): return char == "/"

    @staticmethod
    def ismultiply(char: str): return char == "*"

    @staticmethod
    def isgreat(char: str): return char == ">"

    @staticmethod
    def isless(char: str): return char == "<"

    @staticmethod
    def isdollor(char: str): return char == "$"

    @staticmethod
    def isleftparentheses(char: str): return char == "("

    @staticmethod
    def isrightparentheses(char: str): return char == ")"

    @staticmethod
    def isassign(char: str): return char == "="

    @staticmethod
    def isexclamationpoint(char: str): return char == "!"

    @staticmethod
    def ist(char: str): return char == "t"

    @staticmethod
    def isr(char: str): return char == "r"

    @staticmethod
    def isu(char: str): return char == "u"

    @staticmethod
    def ise(char: str): return char == "e"

    @staticmethod
    def isf(char: str): return char == "f"

    @staticmethod
    def isa(char: str): return char == "a"

    @staticmethod
    def isl(char: str): return char == "l"

    @staticmethod
    def iss(char: str): return char == "s"

    @staticmethod
    def ise(char: str): return char == "e"

    @staticmethod
    def isdot(char: str): return char == "."

    @staticmethod
    def iscomma(char: str): return char == ","

    @staticmethod
    def isstar(char: str): return char == "*"

    @staticmethod
    def isblank(char: str): return char.isspace()

    @staticmethod
    def isand(char: str): return char == "&"

    @staticmethod
    def isL(char: str): return char == "|"

    @staticmethod
    def isquestion(char: str): return char == "?"

    @staticmethod
    def isdoublequotation(char): return char == "\""

    @staticmethod
    def issinglequotation(char): return char == "\'"

    @staticmethod
    def isseparators(char):
        return Character.isblank(char) or \
               Character.isrightparentheses(char) or \
               Character.isand(char) or \
               Character.isL(char) or Character.iscomma(char)

    @staticmethod
    def iskeyword(char):
        """
        是否是保留字段的开头
        :param char:
        :type char:
        :return:
        :rtype:
        """
        return Character.ist(char) or Character.isf(char)

    @staticmethod
    def isoperate(char):
        """
        是否是操作符或运算符
        :param char:
        :type char:
        :return:
        :rtype:
        """
        return char in ["+", "-", "*", "/", "%", "&", "|", "=", ">", "<", "!"]



import enum


@enum.unique
class DfaState(enum.Enum):
    Initial = 0
    Identifier = 1
    Assignment = 2
    Eq = 3
    NEq = 4
    Lt = 5
    Le = 6
    Gt = 7
    Ge = 8
    Var = 9
    Lp = 10
    Rp = 11
    Dq = 12
    Sq = 13
    Colon = 14
    Plus = 15
    Minus = 16
    RealNumber = 17
    ByteAnd = 18
    And = 19
    ByteOr = 20
    Or = 21
    True1 = 22
    True2 = 23
    True3 = 24
    True4 = 25
    False1 = 26
    False2 = 27
    False3 = 28
    False4 = 29
    False5 = 30
    Blank = 31
    DPlus = 32
    PlusEq = 33
    DMinus = 34
    MinusEq = 35
    Comma = 36
    Question = 37
    Ep = 38
    Nidentity = 39
    AwaysEq = 40
    EqMinus = 41
    EqPlus = 42
    Divide = 43
    Multipy = 44
    Dot = 45







from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token
# from edkrule.tests.interpreter.interpreter import Stack

from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser
from edkrule.utils.stack import Stack


# from pybiz.tests.interpreter.parse.parser import Parser


class FunctionStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.bracket_stack = Stack()
        self.statement = Statement()
        self.body_content = []
        self.bodies_content = []

    def accept(self):

        if self.token().type == TokenType.Identifier and self.parser.index < len(
                self.parser.body) - 1 and self.token(offset=1).type == TokenType.Lp:
            # 添加方法名节点
            self.statement.add(self.token())
            # 移到左括号
            self.move()
            # 括号入栈
            self.bracket_stack.push(self.token())
            # 添加左括号节点
            self.statement.add(self.token())
            # 方法体的读取
            while self.bracket_stack.count() != 0:
                self.move()
                # 新的左括号入栈，右括号出栈，保证方法体能读取完
                if self.token().type == TokenType.Rp:
                    self.bracket_stack.pop()
                elif self.token().type == TokenType.Lp:
                    self.bracket_stack.push(self.token())
                # 判断是否是方法体结束的右括号
                # 否: 则加入body_content，以便读取完后再次解析
                if not self.bracket_stack.count() == 0:
                    if self.token().type == TokenType.Comma and self.bracket_stack.count() == 1:
                        self.bodies_content.append(self.body_content.copy())
                        self.body_content.clear()
                    else:
                        self.body_content.append(self.token())
            self.bodies_content.append(self.body_content)
            from edkrule.interpreter.parse.parser import Parser
            for i, body in enumerate(self.bodies_content):
                if len(body) == 0: continue
                parser = Parser(body).parse()
                self.statement.add(parser.statement)
                if i != len(self.bodies_content) - 1:
                    comma = Token(TokenType.Comma, ",")
                    self.statement.add(comma)
            self.statement.add(self.token())
            self.move()
            if not self.statement.empty():
                self.parser.statement.add(self.statement)
            return True

        return False



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.false_token import FalseToken

from edkrule.interpreter.lexer.tokens.true_token import TrueToken


class IdentifierToken:
    @staticmethod
    def accept(lexer, char: str):
        if Character.isalpha(char):
            lexer.state = DfaState.Identifier
            lexer.token_type = TokenType.Identifier
            for t in [TrueToken, FalseToken]:
                if t.accept(lexer, char): return True

            lexer.token_text += char
            return True
        return False



from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser


class EndStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.statement = Statement()

    def accept(self):

        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Comma:
                if self.ternary_end():
                    self.parser.statement.add(self.token())
                    self.move()
                    return True
                else: self.parser.statement.add(self.token())
            elif self.token().type == TokenType.Rp:
                # if self.ternary_end():
                #     self.move()
                #     return True
                self.move()
                return True
        else:
            if self.ternary_end():
                return True
        return False

    def ternary_end(self):
        if self.parser.ternary_stack.count() > 0 and self.parser.ternary_stack.top().type == TokenType.Colon and self.parser.ternary_stack.top(
                -1).type == TokenType.Question:
            i = self.parser.statement.count() - 1
            while i > 0:
                if self.parser.statement.body[i].type == TokenType.Colon:
                    for j in range(i + 1, self.parser.statement.count()):
                        self.statement.add(self.parser.statement.body[j])
                    self.parser.statement.remove(i)
                    self.parser.statement.add(self.statement)
                    break
                i -= 1
            if self.parser.index <= len(self.parser.body) - 1:
                self.statement.add(self.token())
            self.parser.ternary_stack.clear()
            return True
        return False



import abc
# from pybiz.tests.interpreter.parse.parser import Parser


class AbcStatementParser(abc.ABC):

    def __init__(self, parser):
        self.parser = parser

    @abc.abstractmethod
    def accept(self) -> bool: ...

    def token(self, offset: int = 0):
        return self.parser.body[self.parser.index + offset]

    def move(self, pos: int = 1):
        self.parser.index += pos



class Constant:
    DataType = "datatype"
    DataTypes = "datatypes"
    Expressions = "expressions"
    Identifiers = "identifiers"
    Identifier = "identifier"
    Package = "package"
    Class = "class"
    Name = "name"
    DefaultIdentifier = "identifier_jdefault_0063"
    Parent = "parent"




from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser


class ColonStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.statement = Statement()

    def accept(self):
        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Colon:
                if not self.parser.ternary_stack.empty() and self.parser.ternary_stack.top().type == TokenType.Question:
                    i = self.parser.statement.count()-1
                    while i > 0:
                        if self.parser.statement.body[i].type == TokenType.Question:
                            for j in range(i+1, self.parser.statement.count()):
                                self.statement.add(self.parser.statement.body[j])
                            self.parser.statement.remove(i)
                            self.parser.statement.add(self.statement)
                            break
                        i -= 1
                    self.parser.ternary_stack.push(self.token())
                self.parser.statement.add(self.token())
                self.move()
                return True
        return False



import json
import pathlib
from pathlib import Path

from jinja2 import FileSystemLoader, Environment

from edkrule.utils.output.printer import Printer


class PrinterHtml(Printer):



    def output(self, out_path: Path, data, jinja_template_path: Path):
        file_loader = FileSystemLoader(str(jinja_template_path.parent))
        env = Environment(loader=file_loader)

        template = env.get_template(jinja_template_path.name)

        output = template.render(data=data)

        with open(str(out_path), "w", encoding="UTF-8") as f:
            f.write(output)


from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class AssignState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Assignment and Character.isassign(char):
            lexer.token_text += char
            lexer.state = DfaState.Eq
            return True
        elif state == DfaState.Assignment and Character.isplus(char):
            lexer.token_text += char
            lexer.state = DfaState.EqPlus
            return True
        elif state == DfaState.Assignment and Character.isminus(char):
            lexer.token_text += char
            lexer.state = DfaState.EqMinus
            return True
        elif state == DfaState.Assignment:
            if Character.isseparators(char) or Character.iskeyword(char) or Character.isdigit(
                    char) or Character.isalpha(char):
                lexer.token_type = TokenType.Assignment
                lexer.init_token(char)
            else:
                lexer.token_text += char
                lexer.state = DfaState.Identifier
            return True


        return False
        # if state == DfaState.Assignment:
        #     # lexer.token_text += char
        #     lexer.token_type = TokenType.Assignment
        #     lexer.init_token(char)
        #     return True
        # return False



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState


class IdentifierState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Identifier:
            if Character.isalpha(char) or Character.isdigit(char) or Character.isdot(char):
                lexer.token_text += char
            else:
                lexer.state = lexer.init_token(char)
            return True
        return False



from edkrule.interpreter.expression.bracket import Bracket
from edkrule.interpreter.expression.bracket_enum import BracketEnum
from edkrule.interpreter.expression.categorize import Categorize
from edkrule.interpreter.expression.expression import Expression
from edkrule.interpreter.expression.priority import Priority
from edkrule.interpreter.expression.stack import Stack
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token
from edkrule.interpreter.parse.statement import Statement


class Interpreter:
    def __init__(self):
        self._expression_stack = Stack()
        self._op_stack = Stack()
        self._bracket_stack = Stack()
        self.cache = list()

    def is_op(self, stmt):
        return Categorize.is_op(stmt)

    def is_exp(self, stmt):
        return Categorize.is_exp(stmt)

    def is_stmt(self, stmt):
        return Categorize.is_stmt(stmt)

    def interpret(self, statement: Statement):
        i = 0
        while i < len(statement.body):
            stmt = statement.body[i]
            if self.is_stmt(stmt):
                self._expression_stack.push(Interpreter().interpret(stmt))
            if self.is_exp(stmt):
                self._expression_stack.push(stmt)
            if stmt.type in [TokenType.Lp]:
                pre_stmt = statement.body[i - 1]
                be = BracketEnum.Function if pre_stmt.type == TokenType.Identifier else BracketEnum.Calculate
                self._bracket_stack.push(Bracket(be, stmt))
            if stmt.type in [TokenType.Rp]:
                if self._bracket_stack.count() == 0 or self._bracket_stack.top().token.type != TokenType.Lp:
                    raise Exception(f"{''.join([t.text for t in statement.body[0:i]])} () is not pair")
                sub_expressions = self._expression_stack.pop_until(condition=lambda e: e.type == TokenType.Lp)
                expression = Expression([])
                for sub in sub_expressions:
                    sub_expression = Interpreter().interpret(sub) if sub.type == TokenType.Statement else sub
                    expression.append(sub_expression)
                if self._bracket_stack.top().type == BracketEnum.Function:
                    if self._expression_stack.top() is not None:
                        if self._expression_stack.top().type == TokenType.Identifier:
                            expression.name = self._expression_stack.pop()
                self._expression_stack.push(expression)
                self._bracket_stack.pop()
            if self.is_op(stmt):
                if self._bracket_stack.count() != 0:
                    self._expression_stack.push(stmt)
                else:
                    if self._op_stack.count() == 0 or Priority.gt(stmt, self._op_stack.top()):
                        self._op_stack.push(stmt)
                        if stmt.type == TokenType.Question:
                            i = self.ternary_operator(i, statement.body, stmt)

                    else:
                        expression = self.binary_operator()
                        self._expression_stack.push(expression)
                        self._op_stack.push(stmt)
                        if stmt.type == TokenType.Question:
                            i = self.ternary_operator(i, statement.body, stmt)
            i += 1
        while self._op_stack.count() != 0:
            if self._op_stack.count() > 1 and Priority.gt(self._op_stack.top(), self._op_stack.top(-1)):
                expression = self.binary_operator()
                self._expression_stack.push(expression)
            else:
                expression = self.order_binary_operator()
                self._expression_stack.insert(expression)
        return self._expression_stack.pop()

    def order_binary_operator(self):
        op_refs = self._expression_stack.dequeues(2)
        body = [op_refs[0], self._op_stack.dequeue(), op_refs[1]]
        return Expression(body)

    def binary_operator(self):
        op_refs = self._expression_stack.pops(2)
        body = [op_refs[0], self._op_stack.pop(), op_refs[1]]
        return Expression(body)

    def ternary_operator(self, i: int, content: list, question):
        expression_left = Interpreter().interpret(statement=content[i + 1])
        comma = content[i + 2]
        expression_right = Interpreter().interpret(statement=content[i + 3])
        expression = Expression([])

        # expression.append(self._expression_stack.pop())
        # expression.append(self._op_stack.pop())
        # while not self._op_stack.empty() and Priority.lt(question, self._op_stack.top()):
        #     expression.insert(self._op_stack.pop())
        #     expression.insert(self._expression_stack.pop())
        condition_expression = Expression([])
        condition_expression.append(self._expression_stack.pop())
        expression.append(self._op_stack.pop())
        while not self._op_stack.empty() and Priority.lt(question, self._op_stack.top()):
            condition_expression.insert(self._op_stack.pop())
            condition_expression.insert(self._expression_stack.pop())
        expression.insert(condition_expression)
        # expression.append(self._op_stack.pop())
        expression.append(expression_left)
        expression.append(comma)
        expression.append(expression_right)
        expression.name = Token(t_type=TokenType.Identifier, t_text="TernaryOperator")
        self._expression_stack.push(expression)
        return i + 3



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class RealNumberState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.RealNumber:
            if Character.isdigit(char) or Character.isdot(char):
                lexer.token_text += char
            # elif Character.isseparators(char) or Character.isplus(char) or Character.isminus(char):
            else:
                if Character.isdot(lexer.token_text[-1]):
                    raise Exception(f'{lexer.token_text} is invalid number')
                lexer.token_type = TokenType.RealNumber
                lexer.init_token(char)
            return True
        return False



from io import StringIO

from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.states.and_state import AndState
from edkrule.interpreter.lexer.states.assign_state import AssignState
from edkrule.interpreter.lexer.states.aways_eq_state import AwaysEqState
from edkrule.interpreter.lexer.states.bite_and_state import BiteAndState
from edkrule.interpreter.lexer.states.bite_or_state import BiteOrState
from edkrule.interpreter.lexer.states.blank_state import BlankState
from edkrule.interpreter.lexer.states.colon_state import ColonState
from edkrule.interpreter.lexer.states.comma_state import CommaState
from edkrule.interpreter.lexer.states.d_minus_state import DMinusState
from edkrule.interpreter.lexer.states.d_plus_state import DPlusState
from edkrule.interpreter.lexer.states.divide_state import DivideState
from edkrule.interpreter.lexer.states.dot_state import DotState
from edkrule.interpreter.lexer.states.dq_state import DQState
from edkrule.interpreter.lexer.states.ep_state import EpState
from edkrule.interpreter.lexer.states.eq_minus_state import EqMinusState
from edkrule.interpreter.lexer.states.eq_plus_state import EqPlusState
from edkrule.interpreter.lexer.states.eq_state import EqState
from edkrule.interpreter.lexer.states.false_state import FalseState
from edkrule.interpreter.lexer.states.ge_state import GeState
from edkrule.interpreter.lexer.states.gt_state import GtState
from edkrule.interpreter.lexer.states.identifier_state import IdentifierState
from edkrule.interpreter.lexer.states.initial_state import InitialState
from edkrule.interpreter.lexer.states.le_state import LeState
from edkrule.interpreter.lexer.states.lp_state import LpState
from edkrule.interpreter.lexer.states.lt_state import LtState
from edkrule.interpreter.lexer.states.minus_eq_state import MinusEqState
from edkrule.interpreter.lexer.states.minus_state import MinusState
from edkrule.interpreter.lexer.states.multiply_state import MultiplyState
from edkrule.interpreter.lexer.states.neq_state import NeqState
from edkrule.interpreter.lexer.states.nidentity_state import NidentityState
from edkrule.interpreter.lexer.states.or_state import OrState
from edkrule.interpreter.lexer.states.plus_eq_state import PlusEqState
from edkrule.interpreter.lexer.states.plus_state import PlusState
from edkrule.interpreter.lexer.states.question_state import QuestionState
from edkrule.interpreter.lexer.states.real_number_state import RealNumberState
from edkrule.interpreter.lexer.states.rp_state import RpState
from edkrule.interpreter.lexer.states.sq_state import SQState
from edkrule.interpreter.lexer.states.true_state import TrueState
from edkrule.interpreter.lexer.states.variable_state import VariableState
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.assign_token import AssignToken
from edkrule.interpreter.lexer.tokens.bite_and_token import BiteAndToken
from edkrule.interpreter.lexer.tokens.bite_or_token import BiteOrToken
from edkrule.interpreter.lexer.tokens.blank_token import BlankToken
from edkrule.interpreter.lexer.tokens.colon_token import ColonToken
from edkrule.interpreter.lexer.tokens.comma_token import CommaToken
from edkrule.interpreter.lexer.tokens.divide_token import DivideToken
from edkrule.interpreter.lexer.tokens.dot_token import DotToken
from edkrule.interpreter.lexer.tokens.dq_token import DqToken
from edkrule.interpreter.lexer.tokens.ep_token import EpToken
from edkrule.interpreter.lexer.tokens.false_token import FalseToken
from edkrule.interpreter.lexer.tokens.gt_token import GtToken
from edkrule.interpreter.lexer.tokens.identifier_token import IdentifierToken
from edkrule.interpreter.lexer.tokens.lp_token import LpToken
from edkrule.interpreter.lexer.tokens.lt_token import LtToken
from edkrule.interpreter.lexer.tokens.minus_token import MinusToken
from edkrule.interpreter.lexer.tokens.multiply_token import MultiplyToken
from edkrule.interpreter.lexer.tokens.neq_token import NeqToken
from edkrule.interpreter.lexer.tokens.plus_token import PlusToken
from edkrule.interpreter.lexer.tokens.question_token import QuestionToken
from edkrule.interpreter.lexer.tokens.realnumber_token import RealNumberToken
from edkrule.interpreter.lexer.tokens.rp_token import RpToken
from edkrule.interpreter.lexer.tokens.sq_token import SqToken
from edkrule.interpreter.lexer.tokens.token import Token
from edkrule.interpreter.lexer.tokens.true_token import TrueToken
from edkrule.interpreter.lexer.tokens.variable_token import VariableToken


class Lexer:
    def __init__(self):
        self.token_list = []
        self.token_type = None
        self.token_text = ""
        self.state = DfaState.Initial

        # 为了Variable 的 *.* 方式
        self.dot_numer = 0
        self.next_dot = False

    def reset(self):
        self.token_type = None
        self.token_text = ""

    def clean(self):
        self.reset()
        self.state = DfaState.Initial
        self.token_list = []

    def init_token(self, char: str) -> DfaState:
        if self.token_text:
            token = Token(self.token_type, self.token_text)
            self.token_list.append(token)
            self.reset()

        self.state = DfaState.Initial
        for t in [IdentifierToken, BlankToken, BiteAndToken, BiteOrToken, RealNumberToken,
                  PlusToken, MinusToken, ColonToken, SqToken, DqToken, CommaToken, QuestionToken,
                  LpToken, RpToken, VariableToken, GtToken, LtToken, EpToken, NeqToken, AssignToken,
                  MultiplyToken, DivideToken, DotToken]:
            if t.accept(self, char):
                break
        return self.state

    def tokenize(self, code):
        self.clean()
        while True:
            char = code.read(1)
            if not char:
                self.last(char)
                break
            for s in [InitialState, IdentifierState, TrueState, FalseState, BlankState,
                      BiteAndState, AndState, BiteOrState, OrState, RealNumberState, PlusState,
                      DPlusState, PlusEqState, MinusState, DMinusState, MinusEqState, ColonState,
                      SQState, DQState, CommaState, QuestionState, LpState, RpState, VariableState,
                      GtState, GeState, LtState, LeState, EpState, NeqState, NidentityState, AssignState,
                      EqState, EqPlusState, EqMinusState, AwaysEqState, DivideState, MultiplyState, DotState]:
                if s.accept(self, char, self.state):
                    break

    def last(self, char):
        if self.state == DfaState.True4:
            self.token_type = TokenType.TRUE
        elif self.state == DfaState.Identifier:
            self.token_type = TokenType.Identifier
        elif self.state == DfaState.False5:
            self.token_type = TokenType.FALSE
        elif self.state == DfaState.Blank:
            self.token_type = TokenType.Blank
        elif self.state == DfaState.ByteAnd:
            self.token_type = TokenType.ByteAnd
        elif self.state == DfaState.And:
            self.token_type = TokenType.And
        elif self.state == DfaState.ByteOr:
            self.token_type = TokenType.ByteOr
        elif self.state == DfaState.Or:
            self.token_type = TokenType.Or
        elif self.state == DfaState.RealNumber:
            self.token_type = TokenType.RealNumber
        elif self.state == DfaState.Plus:
            self.token_type = TokenType.Plus
        elif self.state == DfaState.DPlus:
            self.token_type = TokenType.DPlus
        elif self.state == DfaState.PlusEq:
            self.token_type = TokenType.PlusEq
        elif self.state == DfaState.Minus:
            self.token_type = TokenType.Minus
        elif self.state == DfaState.DMinus:
            self.token_type = TokenType.DMinus
        elif self.state == DfaState.MinusEq:
            self.token_type = TokenType.MinusEq
        elif self.state == DfaState.Colon:
            self.token_type = TokenType.Colon
        elif self.state == DfaState.Comma:
            self.token_type = TokenType.Comma
        elif self.state == DfaState.Question:
            self.token_type = TokenType.Question
        elif self.state == DfaState.Lp:
            self.token_type = TokenType.Lp
        elif self.state == DfaState.Rp:
            self.token_type = TokenType.Rp
        elif self.state == DfaState.Var:
            self.token_type = TokenType.Variable
            self.dot_numer = 0
            self.next_dot = False
        if self.state != DfaState.Initial:
            self.init_token(char)

    def display(self):
        print([(t.type, t.text) for t in self.token_list])


if __name__ == '__main__':
    lexer = Lexer()
    # string = StringIO("()")
    #
    # string = StringIO("(1+2)")
    #
    string = StringIO(
"return mustAnswer([7])")
    #
    # string = StringIO("('ass' + '1+2')")
    # string = StringIO("a!=''")

    # string = StringIO("1!==2")

    lexer.tokenize(string)
    for t in lexer.token_list:
        print(t.type, t.text)



from edkrule.engine.finder.finder_by_rid import FinderByRid
from edkrule.engine.finder.finder_by_rule_fragment import FinderByRuleFragment
from edkrule.utils.format.blank_formater import BlankFormater


class By:
    Rid = FinderByRid()
    RuleFragment = FinderByRuleFragment(BlankFormater())



class Cache:
    def __init__(self):
        self._data = {}

    def get(self, key, default=None):
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value



import json
from pathlib import Path

from edkrule.edk_rule import EdkRule

# expression = EdkRule().expression("condition(mustAnswer(toNum($*.*.*)), isRange($*.*.*, 40,160))")
# expression = EdkRule().expression("toDate($*.*.*)!=\"\"&&toDate(getMinByLog($*.Informed Consent.DSSTDAT))!=\"\"?dateDiff(getMinByLog($*.Informed Consent.DSSTDAT),$*.*.*, \"D\")>0:true")
# expression = EdkRule().expression("""autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)""")
expression = EdkRule().draw("""autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)""", Path("test.html"))
# data = expression.tree_data()
# print(data)
# with open("tree.json", 'w') as f:
#     json.dump(data, f)



from edkrule.interpreter.lexer.token_type import TokenType
# from edkrule.tests.interpreter.interpreter import Stack

from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser
from edkrule.utils.stack import Stack


# from pybiz.tests.interpreter.parse.parser import Parser


class BracketStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.bracket_stack = Stack()
        self.statement = Statement()
        self.body_content = []

    def accept(self):
        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Lp:
                # 括号入栈
                self.bracket_stack.push(self.token())
                # 添加左括号节点
                self.statement.add(self.token())
                # 方法体的读取
                while self.bracket_stack.count() != 0:
                    self.move()
                    # 新的左括号入栈，右括号出栈，保证方法体能读取完
                    if self.token().type == TokenType.Rp:
                        self.bracket_stack.pop()
                    elif self.token().type == TokenType.Lp:
                        self.bracket_stack.push(self.token())
                    # 判断是否是方法体结束的右括号
                    # 否: 则加入body_content，以便读取完后再次解析
                    if not self.bracket_stack.count() == 0:
                        self.body_content.append(self.token())
                from edkrule.interpreter.parse.parser import Parser
                parser = Parser(self.body_content).parse()

                self.statement.add(parser.statement)
                self.statement.add(self.token())
                self.move()
                if not self.statement.empty():
                    self.parser.statement.add(self.statement)
                return True
        return False



class TernaryStack:
    def __init__(self):
        self.stack = []

    def append(self, e):
        self.stack.append([e])

    def pop(self, index=-1):
        self.stack[index].pop()
        if len(self.stack[index]) == 0: self.stack.pop()

    def empty(self) -> bool:
        return len(self.stack) == 0



from edkrule.utils.cache import Cache
from edkrule.engine.engine_definition import EngineDefinition
from edkrule.engine.sys.sys_path import sys_path
from edkrule.engine.xml.xml_loader import XMLLoader
from edkrule.utils.squeue import Squeue


class Engine:
    def __init__(self, definition_path, use_sys=True):
        self._cache = Cache()
        self._squeue = Squeue()
        self._definition_path = definition_path
        self._engine_definition = EngineDefinition()
        self.re_loader(definition_path=self._definition_path, use_sys=use_sys)

    @property
    def cache(self): return self._cache

    @property
    def squeue(self): return self._squeue

    def re_loader(self, definition_path, use_sys=True, clear=True):
        """
        传参方式 1.
            definition_path = “” and use_sys=True 只是用系统默认规则引擎
        传参方式 2.
            definition_path = “xxxxx” and use_sys=True 同时加载传入的引擎配置和系统默认规则引擎
            如果有相同的名字的，加载传入的引擎配置会覆盖系统默认的配置
        传参方式 3.
            definition_path = “xxxxx” and use_sys=False 只加载传入的引擎配置
        加载 引擎的 XML 配置
        :param definition_path: 加载的 引擎配置的 XML 路径
        :type definition_path:
        :param use_sys: 是否加载系统默认的规则
        :type use_sys:
        :param clear: 重新加载的时候，是否清楚原配置
        :type clear:
        :return:
        :rtype:
        """
        if clear: self._engine_definition = EngineDefinition()
        if use_sys: XMLLoader(sys_path()).loader(engine_definition=self._engine_definition)
        if len(definition_path) != 0:
            XMLLoader(definition_path).loader(engine_definition=self._engine_definition)

    def get(self, definition_type, name: str):
        return self._engine_definition.get(definition_type, name.lower())

    def get_class(self, definition_type, name: str):
        return self._engine_definition.get_class(definition_type, name.lower())

    @property
    def definition(self):
        return self._engine_definition

    # def get(self, key):
    #     return self._engine_definition.get(key)
